@@ -1,5 +1,5 @@
 /* dn.c - Distinguished Name helper functions
- *      Copyright (C) 2001 g10 Code GmbH
+ *      Copyright (C) 2001, 2006 g10 Code GmbH
  *
  * This file is part of KSBA.
  *
@@ -15,7 +15,8 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
+ * USA.
  */
 
 /* Reference is RFC-2253 */
@@ -169,6 +170,43 @@ put_stringbuf_mem (struct stringbuf *sb, const char *text, size_t n)
   sb->len += n;
 }
 
+static void
+put_stringbuf_mem_skip (struct stringbuf *sb, const char *text, size_t n,
+                        int skip)
+{
+  char *p;
+      
+  if (!skip)
+    {
+      put_stringbuf_mem (sb, text, n);
+      return;
+    }
+  if (sb->out_of_core)
+    return;
+
+  if (sb->len + n >= sb->size)
+    {
+      /* Note: we allocate too much here, but we don't care. */
+      sb->size += n + 100;
+      p = xtryrealloc (sb->buf, sb->size);
+      if ( !p)
+        {
+          sb->out_of_core = 1;
+          return;
+        }
+      sb->buf = p;
+    }
+  p = sb->buf+sb->len;
+  while (n > skip)
+    {
+      text += skip;
+      n -= skip;
+      *p++ = *text++;
+      n--;
+      sb->len++;
+    }
+}
+
 static char *
 get_stringbuf (struct stringbuf *sb)
 {
@@ -191,9 +229,11 @@ get_stringbuf (struct stringbuf *sb)
 /* This function is used for 1 byte encodings to insert any required
    quoting.  It does not do the quoting for a space or hash mark at
    the beginning of a string or a space as the last character of a
-   string */
+   string.  It will do steps of SKIP+1 characters, assuming that these
+   SKIP characters are null octets. */
 static void
-append_quoted (struct stringbuf *sb, const unsigned char *value, size_t length)
+append_quoted (struct stringbuf *sb, const unsigned char *value, size_t length,
+               int skip)
 {
   unsigned char tmp[4];
   const unsigned char *s = value;
@@ -201,16 +241,20 @@ append_quoted (struct stringbuf *sb, const unsigned char *value, size_t length)
 
   for (;;)
     {
-      for (value = s; n < length; n++, s++)
+      for (value = s; n+skip < length; n++, s++)
         {
+          s += skip;
+          n += skip;
           if (*s < ' ' || *s > 126 || strchr (",+\"\\<>;", *s) )
             break;
         }
 
       if (s != value)
-        put_stringbuf_mem (sb, value, s-value);
-      if (n==length)
+        put_stringbuf_mem_skip (sb, value, s-value, skip);
+      if (n+skip >= length)
         return; /* ready */
+      s += skip;
+      n += skip;
       if ( *s < ' ' || *s > 126 )
         {
           sprintf (tmp, "\\%02X", *s);
@@ -259,7 +303,7 @@ append_utf8_value (const unsigned char *value, size_t length,
       for (value = s; n < length && !(*s & 0x80); n++, s++)
         ;
       if (s != value)
-        append_quoted (sb, value, s-value);
+        append_quoted (sb, value, s-value, 0);
       if (n==length)
         return; /* ready */
       assert ((*s & 0x80));
@@ -322,7 +366,7 @@ append_latin1_value (const unsigned char *value, size_t length,
       for (value = s; n < length && !(*s & 0x80); n++, s++)
         ;
       if (s != value)
-        append_quoted (sb, value, s-value);
+        append_quoted (sb, value, s-value, 0);
       if (n==length)
         return; /* ready */
       assert ((*s & 0x80));
@@ -368,7 +412,7 @@ append_ucs4_value (const unsigned char *value, size_t length,
              && !s[0] && !s[1] && !s[2] && !(s[3] & 0x80); n += 4, s += 4)
         ;
       if (s != value)
-        append_quoted (sb, value, s-value);
+        append_quoted (sb, value, s-value, 3);
       if (n>=length)
         return; /* ready */
       if (n < 4)
@@ -442,7 +486,7 @@ append_ucs2_value (const unsigned char *value, size_t length,
       value += 2;
       length -= 2;
     }
-  if (length>3 && !value[0] && value[1] == ' ')
+  if (length>1 && !value[0] && value[1] == ' ')
     {
       tmp[0] = '\\';
       tmp[1] = ' ';
@@ -455,7 +499,7 @@ append_ucs2_value (const unsigned char *value, size_t length,
       for (value = s; n+1 < length && !s[0] && !(s[1] & 0x80); n += 2, s += 2)
         ;
       if (s != value)
-        append_quoted (sb, value, s-value);
+        append_quoted (sb, value, s-value, 1);
       if (n>=length)
         return; /* ready */
       if (n < 2)
@@ -842,13 +886,16 @@ write_escaped (ksba_writer_t w, const unsigned char *buffer, size_t nbytes)
 
 
 /* Parse one RDN, and write it to WRITER.  Returns a pointer to the
-   next RDN part where the comma has alrady been skipped or NULL in
-   case of an error.  When NULL is passed as WRITER, the fucntion does
+   next RDN part where the comma has already been skipped or NULL in
+   case of an error.  When NULL is passed as WRITER, the function does
    not allocate any memory but just parses the string and returns the
-   ENDP. */
+   ENDP. If ROFF or RLEN are not NULL, they will receive informaion
+   useful for error reporting. */
 static gpg_error_t
-parse_rdn (const unsigned char *string, const char **endp, ksba_writer_t writer)
+parse_rdn (const unsigned char *string, const char **endp, 
+           ksba_writer_t writer, size_t *roff, size_t *rlen)
 {
+  const unsigned char *orig_string = string;
   const unsigned char *s, *s1;
   size_t n, n1;
   int i;
@@ -862,15 +909,24 @@ parse_rdn (const unsigned char *string, const char **endp, ksba_writer_t writer)
   int valuetype;
   int need_escaping = 0;
   gpg_error_t err = 0;
+  size_t dummy_roff, dummy_rlen;
+
+  if (!roff)
+    roff = &dummy_roff;
+  if (!rlen)
+    rlen = &dummy_rlen;
+
+  *roff = *rlen = 0;
 
   if (!string)
     return gpg_error (GPG_ERR_INV_VALUE);
   while (*string == ' ')
     string++;
+  *roff = string - orig_string;
   if (!*string)
     return gpg_error (GPG_ERR_SYNTAX); /* empty elements are not allowed */
   s = string;
-
+  
   if ( ((*s == 'o' && s[1] == 'i' && s[2] == 'd')
         ||(*s == 'O' && s[1] == 'I' && s[2] == 'D'))
        && s[3] == '.' && digitp (s+4))
@@ -878,6 +934,7 @@ parse_rdn (const unsigned char *string, const char **endp, ksba_writer_t writer)
 
   /* parse attributeType */
   string = s;
+  *roff = string - orig_string;
   if (digitp (s))
     { /* oid */
       for (s++; digitp (s) || (*s == '.' && s[1] != '.'); s++)
@@ -920,7 +977,11 @@ parse_rdn (const unsigned char *string, const char **endp, ksba_writer_t writer)
             break;
         }
       if (!oid_name_tbl[i].name)
-        return gpg_error (GPG_ERR_UNKNOWN_NAME);
+        {
+          *roff = string - orig_string;
+          *rlen = n;
+          return gpg_error (GPG_ERR_UNKNOWN_NAME);
+        }
       oid = oid_name_tbl[i].oid;
       oidlen = oid_name_tbl[i].oidlen;
     }
@@ -929,6 +990,8 @@ parse_rdn (const unsigned char *string, const char **endp, ksba_writer_t writer)
     s++;
   string = s;
 
+  *roff = string - orig_string;
+
   /* parse attributeValue */
   if (!*s)
     {
@@ -947,6 +1010,7 @@ parse_rdn (const unsigned char *string, const char **endp, ksba_writer_t writer)
       n = s - string;
       if (!n || (n & 1))
         {
+          *rlen = n;
           err = gpg_error (GPG_ERR_SYNTAX); /* no hex digits or odd number */
           goto leave;
         }
@@ -994,6 +1058,7 @@ parse_rdn (const unsigned char *string, const char **endp, ksba_writer_t writer)
       s = count_quoted_string (string, &n, 1, &valuetype);
       if (!s || *s != '\"')
         {
+          *rlen = s - orig_string;
           err = gpg_error (GPG_ERR_SYNTAX); /* error or quote not closed */
           goto leave;
         }
@@ -1026,11 +1091,14 @@ parse_rdn (const unsigned char *string, const char **endp, ksba_writer_t writer)
     }
   if ( *s && *s != ',' && *s != ';' && *s != '+')
     {
+      *roff = s - orig_string;
       err = gpg_error (GPG_ERR_SYNTAX); /* invalid delimiter */
       goto leave;
     }
   if (*s == '+') /* fixme: implement this */
     {
+      *roff = s - orig_string;
+      *rlen = 1;
       err = gpg_error (GPG_ERR_NOT_IMPLEMENTED); 
       goto leave;
     }
@@ -1093,18 +1161,18 @@ _ksba_dn_from_str (const char *string, char **rbuf, size_t *rlength)
   int part_array_size, nparts;
 
   *rbuf = NULL; *rlength = 0;
-  /* We are going to build the object using a writer object */
+  /* We are going to build the object using a writer object.  */
   err = ksba_writer_new (&writer);
   if (!err)
     err = ksba_writer_set_mem (writer, 1024);
   if (err)
     return err;
 
-  /* We must assign it in reverese order so we do it in 2 passes */
+  /* We must assign it in reverse order so we do it in 2 passes. */
   part_array_size = 0;
   for (nparts=0, s=string; s && *s;)
     {
-      err = parse_rdn (s, &endp, NULL);
+      err = parse_rdn (s, &endp, NULL, NULL, NULL);
       if (err)
         goto leave;
       if (nparts >= part_array_size)
@@ -1133,19 +1201,19 @@ _ksba_dn_from_str (const char *string, char **rbuf, size_t *rlength)
 
   while (--nparts >= 0)
     {
-      err = parse_rdn (part_array[nparts], &endp, writer);
+      err = parse_rdn (part_array[nparts], &endp, writer, NULL, NULL);
       if (err)
         goto leave;
     }
 
-  /* Now get the memory */
+  /* Now get the memory.  */
   buf = ksba_writer_snatch_mem (writer, &buflen);
   if (!buf)
     {
       err = gpg_error (GPG_ERR_ENOMEM);
       goto leave;
     }
-  /* reinitialize the buffer to create the outer sequence*/
+  /* Reinitialize the buffer to create the outer sequence.  */
   err = ksba_writer_set_mem (writer, buflen + 10);
   if (err)
     goto leave;
@@ -1176,3 +1244,60 @@ _ksba_dn_from_str (const char *string, char **rbuf, size_t *rlength)
 }
 
 
+
+gpg_error_t
+ksba_dn_der2str (const void *der, size_t derlen, char **rstring)
+{
+  return _ksba_derdn_to_str (der, derlen, rstring);
+}
+
+
+gpg_error_t
+ksba_dn_str2der (const char *string, unsigned char **rder, size_t *rderlen)
+{
+  return _ksba_dn_from_str (string, (char**)rder, rderlen);
+}
+
+
+
+/* Assuming that STRING contains an rfc2253 encoded string, test
+   whther this string may be passed as a valid DN to libksba.  On
+   success the functions returns 0.  On error the function returns an
+   error code and stores the offset within STRING of the erroneous
+   part at RERROFF. RERRLEN will then receive the length of the
+   erroneous part.  This function is most useful to test whether a
+   symbolic name (like SN) is supported. SEQ should be passed as 0 for
+   now.  RERROFF and RERRLEN may be passed as NULL if the caller is
+   not interested at this value. */
+gpg_error_t 
+ksba_dn_teststr (const char *string, int seq, 
+                 size_t *rerroff, size_t *rerrlen)
+{
+  size_t dummy_erroff, dummy_errlen;
+  gpg_error_t err;
+  int nparts;
+  const char *s, *endp;
+  size_t off, len;
+
+  if (!rerroff)
+    rerroff = &dummy_erroff;
+  if (!rerrlen)
+    rerrlen = &dummy_errlen;
+
+  *rerrlen = *rerroff = 0;
+
+  for (nparts=0, s=string; s && *s; nparts++)
+    {
+      err = parse_rdn (s, &endp, NULL, &off, &len);
+      if (err && !seq--)
+        {
+          *rerroff = s - string + off;
+          *rerrlen = len? len : strlen (s);
+          return err;
+        }
+      s = endp;
+    }
+  if (!nparts)
+    return gpg_error (GPG_ERR_SYNTAX);
+  return 0;
+}
